#!/bin/env python3
#
# This example reads a test pulse from the Timepix4 over 10 Gb Ethernet
#
import grpc
import sys
import time

import numpy as np
from spidr4 import rpc, tpx4tools, utils, stream, tpx4regs
import queue
import matplotlib.pyplot as plt


if len(sys.argv) != 3:
  print("./tpx4_xgbe_testpulse_local_daq.py <spidr4_ip> <10gbe network interface>")
  sys.exit(1)

spidr4_ip = sys.argv[1]
iface2find = sys.argv[2]

XGBE_PORT=8192

PATTERN_STRIDE=2

# Find network interface information
# -----------------------------------------------------------------------------------------------------------
# Get network card MAC address and IP address
xgbe_host_mac, xgbe_host_ip = utils.get_nic_info(iface2find)

# make-up an 10gbe IP for the spidr4 module, as long as it is not the same as xgbe_host_ip
xgbe_spidr_ip = utils.inc_ip(xgbe_host_ip)

# Main loop, create network connection
with grpc.insecure_channel(spidr4_ip + ":50051") as channel:
    # Get the services
    # ------------------------------------------------------------------------------------------------------
    ctrl = rpc.ControlInfoStub(channel)
    tpx4 = rpc.Timepix4Stub(channel)
    datastream = rpc.DataStreamStub(channel)

    # Reset the pixel chips (will also load the default configuration)
    # ------------------------------------------------------------------------------------------------------
    ctrl.ResetPixelChips(rpc.EMPTY)

    # Configure the output
    # ------------------------------------------------------------------------------------------------------
    datastream.ConfigXGbe(rpc.XGbeConfig(
        idx=0,
        primary=rpc.XGbeLinkConfig(
            source_ip=xgbe_spidr_ip,
            dest_ip=xgbe_host_ip,
            dest_mac=xgbe_host_mac,
            port=XGBE_PORT
        )
    ))

    # Configure the test pulse
    # ------------------------------------------------------------------------------------------------------
    tpConfig = rpc.Tpx4TestPulseConfig(
        idx=0,
        count=1,
        period_on_us=1,
        period_off_us=10000,
        digital=True,
        link_shutter=True
    )
    tpx4.TestPulseSetConfig(tpConfig)

    # Configure event-based readout
    # ------------------------------------------------------------------------------------------------------
    readoutCfg = rpc.Tpx4ReadoutConfig(
        idx=0,
        mode=rpc.TPX4_READOUT_TOA_TOT,
        pc24b_thr=1
    )
    tpx4.ReadoutSetConfig(readoutCfg)

    # Configure Pixel Matrix
    # ------------------------------------------------------------------------------------------------------
    pixel_cfg = tpx4tools.PixelConfig(dac=31, power_enable=True, tp_enable=False, mask=False).word
    pixel_cfg_tp = tpx4tools.PixelConfig(dac=31, power_enable=True, tp_enable=True, mask=False).word
    spixel_cfg = np.full((512, 448), pixel_cfg, dtype=np.uint8)

    img = np.load("timepix_test_image.npz")['img']
    for y in range(0, 512):
        for x in range(0, 448):
            if img[y,x] and ((x+y) % PATTERN_STRIDE == 0):
                spixel_cfg[y, x] = pixel_cfg_tp

    config_blob = tpx4tools.logic2chip_cfg_matrix(spixel_cfg)

    tpx4.ConfigPixels(
        rpc.Tpx4PixelConfig(
            idx=0,
            config=config_blob.tobytes()
        )
    )

    # Configure shutter
    # ------------------------------------------------------------------------------------------------------
    tpx4.ShutterSetConfig(
        rpc.Tpx4ShutterConfig(
            idx=0,
            mode=rpc.TPX4_SHUTTER_MODE_PROG_SINGLE,
            input=rpc.TPX4_SHUTTER_INPUT_SLOW_CONTROL,
            prog_open_us=30000,
            prog_close_us=20
        )
    )

    tpx4.SimpleWrite(rpc.SimpleWriteRequest(idx=0, addr=tpx4regs.MATRIX_CTRL, val=0xA))


    # Configure local data acquisition
    # ------------------------------------------------------------------------------------------------------
    q = queue.Queue()
    prt1 = stream.udp_read_thread(xgbe_host_ip, XGBE_PORT, q) 
    prt2 = stream.udp_read_thread(xgbe_host_ip, XGBE_PORT + 1, q) 

    # Setup verification matrix
    # ------------------------------------------------------------------------------------------------------
    img2 = np.zeros(shape=(512, 448), dtype=int)


    # Enable the test-pulse
    # ------------------------------------------------------------------------------------------------------
    tpx4.TestPulseEnable(rpc.ChipIndex(idx=0))

    # Test-pulse start
    # ------------------------------------------------------------------------------------------------------

    print("Starting test-pulse")
    tpx4.TestPulseStart(rpc.ChipIndex(idx=0))


    to = time.time() + 10
    # Read-data and put in matrix
    # ------------------------------------------------------------------------------------------------------
    for data in stream.queue_generator(q, 5):
      if time.time() > to:
        break
      pkt = tpx4tools.decode_dd_packet(data)
      if isinstance(pkt, tpx4tools.ToAToTPacket):
        x, y = tpx4tools.chip_coords2logic(pkt.Top, pkt.EoC, pkt.SPGroup, pkt.SPixel, pkt.Pixel)
        img2[y,x] += 1

    print("Wait complete, stopping threads")
    prt1.stop()
    prt2.stop()
    
    # Disable data acquisition.
    # ------------------------------------------------------------------------------------------------------
    datastream.ConfigNone(rpc.ChipIndex(idx=0))

    error = np.zeros(shape=(512, 448), dtype=int)
    for y in range(0, 512):
        for x in range(0, 448):
          hit = 1 if img[y,x] and ((x+y) % PATTERN_STRIDE == 0) else 0
          if hit != img2[y, x]:
            error[y, x] = 1

    
    fig, (ax1, ax2) = plt.subplots(ncols=2)
    fig.set_size_inches(18,9)
    ax1.imshow(np.clip(img2, 0, 2), origin='lower')
    ax2.imshow(np.clip(error, 0, 1), origin='lower')
    plt.tight_layout()
